// ENGR 190 final project, reaction timer test bench
// Rick Nungester 3/15/17

`timescale 1ms / 10us      // time in ms units, to 0.01 ms resolution

module  ReactionTimer_tb;
                           // DUT inputs
  reg        clk_50 = 1;   // simulated 50 MHz clock
  reg        nRst;         // active-low user reset event
  reg        nReact;       // active-low user reaction event
  reg  [6:0] armT;         // 6:4 10s digit, 3:0 1s digit

                           // DUT outputs
  wire       react_st;     // on in React state
  wire [4:0] cnt10s;       // {blank, 4-bit value} for 5 digits
  wire [4:0] cnt1s;
  wire [4:0] cnt100ms;
  wire [4:0] cnt10ms;
  wire [4:0] cnt1ms;

  ReactionTimer    // instantiate the Device Under Test
    #(.tc0_p (1))  // 2 clocks/ms for simulation
    DUT (
      .clk_i       (clk_50),
      .nRst_i      (nRst),
      .nReact_i    (nReact),
      .armT_i      (armT),
      .reactSt_o   (react_st),
      .cnt10s_o    (cnt10s),
      .cnt1s_o     (cnt1s),
      .cnt100ms_o  (cnt100ms),
      .cnt10ms_o   (cnt10ms),
      .cnt1ms_o    (cnt1ms) );

  always
    #0.25 clk_50 = ~clk_50;  // 2 clocks/ms, +edges at 0.5, 1, 1.5... ms

  initial begin            // test procedure
    $dumpfile("a.vcd");              // for GTKWave
    $dumpvars(0, ReactionTimer_tb);  // for GTKWave
    $display("Reaction Timer Simulation: Start");

            nRst   = 1'd1;  // @ t = 0 initialize inputs: not reset,
            nReact = 1'b1;  // not reacting,
            armT   = 7'h01; // arm period = 1s

                            // SIMPLE TEST CASE
    #50     nRst   = 1'b0;  // @ 50.0 ms press the reset key
    #50     nRst   = 1'b1;  // @ 100.0 ms release the reset key
    #1123.4 nReact = 1'b0;  // @ 1223.4 ms react (1s arm + 123.4 ms react)
    #50     nReact = 1'b1;  // @ 1273.4 ms release react
    #50                     // @ 1323.4 ms finish
    $display("Reaction Timer Simulation: End");
    $finish;

    //                        // COMPLICATED TEST CASE
    // #50    nRst   = 1'b0;  // @ 50.0 ms enter the Reset state
    // #49.9  nRst   = 1'b1;  // @ 99.9 ms release the reset key
    // #0.1                   // @ 100.0 ms (clock +edge) Reset -> Arming
    // #1000                  // @ 1100.0 ms timer gets to 1.000s
    // #0.5                   // @ 1100.5 ms (+1 clk), React state, timer = 0
    // #0.5                   // @ 1101.0 ms (+1 clk) release reset, timing...
    // #123.4 nReact = 1'b0;  // @ 1224.4 ms (123.4 ms reaction time), react
    // #0.1                   // @ 1224.5 ms (+edge) DISPLAY STATE, stop timer
    // #25.5  nReact = 1'b1;  // @ 1250.0 ms release the react key
    // #50                    // @ 1300.0 ms end simulation
    // $display("Reaction Timer Simulation: End");
    // $finish;
  end

endmodule


// ENGR 190 lab 10, Reaction Timer on the DE2-115 board
// Rick Nungester 3/12/17

module ReactionTimer_top (
  input        CLOCK_50, // 50 MHz clock
  input  [6:0] SW,       // hex 00 to 59 second BCD arm period
  input        KEY3,     // press for React -> Display transition
  input        KEY0,     // press to Reset, release to Arming
  output       LEDR0,    // "react now" LED
  output [6:0] HEX5,     // Timer 7-seg displays (00.000 to 99.999 s)
  output [6:0] HEX4,
  output [6:0] HEX3,
  output [6:0] HEX2,
  output [6:0] HEX1 );

  wire [4:0] cnt10s_w;  // output digits with blanking
  wire [4:0] cnt1s_w;
  wire [4:0] cnt100ms_w;
  wire [4:0] cnt10ms_w;
  wire [4:0] cnt1ms_w;

  ReactionTimer
    //#(.tc0_p       (28934),    // 12^-3 second update rate
    //  .tc1_p       (11) )      // base 12 timing
    //#(.tc0_p       (6249999),  // 2^-3 second update rate
    //  .tc1_p       (1) )       // base 2 timing
    RT (
      .clk_i      (CLOCK_50),
      .nRst_i     (KEY0),
      .nReact_i   (KEY3),
      .armT_i     (SW),
      .reactSt_o  (LEDR0),
      .cnt10s_o   (cnt10s_w),
      .cnt1s_o    (cnt1s_w),
      .cnt100ms_o (cnt100ms_w),
      .cnt10ms_o  (cnt10ms_w),
      .cnt1ms_o   (cnt1ms_w) );

  hexBlank7seg disp10s (
    .bn_i   (cnt10s_w),
    .disp_o (HEX5) );

  hexBlank7seg disp1s (
    .bn_i   (cnt1s_w),
    .disp_o (HEX4) );

  hexBlank7seg disp100ms (
    .bn_i   (cnt100ms_w),
    .disp_o (HEX3) );

  hexBlank7seg disp10ms (
    .bn_i   (cnt10ms_w),
    .disp_o (HEX2) );

  hexBlank7seg disp1ms (
    .bn_i   (cnt1ms_w),
    .disp_o (HEX1) );

endmodule


// ENGR 190 lab 10, Reaction Timer (see the assignment for requirements)
// Rick Nungester 3/17/17

module ReactionTimer #(
  parameter    tc0_p = 49999,  // 1 ms update period (see Cnt5Digits)
  parameter    tc1_p = 9 ) (   // time in base 10 (see Cnt5Digits)
  input        clk_i,          // 50 MHz clock
  input        nRst_i,         // active low asynchronous reset
  input        nReact_i,       // active low react key
  input  [6:0] armT_i,         // arm period (7'h00-7f, to BCD 00-59 sec)
  output       reactSt_o,      // high when in the React state
  output [4:0] cnt10s_o,       // 5 timer digits with blanking control
  output [4:0] cnt1s_o,
  output [4:0] cnt100ms_o,
  output [4:0] cnt10ms_o,
  output [4:0] cnt1ms_o );

  localparam resetSt_p = 4'b0001;  // one-hot state assignment
  localparam armSt_p   = 4'b0010;
  localparam reactSt_p = 4'b0100;
  localparam dispSt_p  = 4'b1000;

  wire [3:0] cnt10s_w;    // timer output digits
  wire [3:0] cnt1s_w;
  wire [3:0] cnt100ms_w;
  wire [3:0] cnt10ms_w;
  wire [3:0] cnt1ms_w;

  wire nRst_timer_w;      // timer reset & enable controls
  wire en_timer_w;

  reg  [3:0] armT10s_r;   // BCD arm period (00-59s) at reset-to-arm
  reg  [3:0] armT1s_r;

  reg [3:0] prevSt_r;     // previous, current, and next state
  reg [3:0] currSt_r;
  reg [3:0] nextSt_r;

  reg prev_nReact_r;      // prev "react" key (for transition sensing)

  // Define the next state combinational circuit.
  always @(currSt_r, nRst_i, nReact_i, prev_nReact_r,
           cnt10s_w, cnt1s_w, armT10s_r, armT1s_r)
    case (currSt_r)
      resetSt_p: nextSt_r = ((nRst_i) ? armSt_p : resetSt_p);
        armSt_p: nextSt_r = (  (  (cnt10s_w == armT10s_r)
                                & (cnt1s_w == armT1s_r) )
                             ? reactSt_p : armSt_p);
      reactSt_p: nextSt_r = (  (prev_nReact_r & !nReact_i)
                             ? dispSt_p : reactSt_p);
       dispSt_p: nextSt_r = dispSt_p;  // see !nRst_i below
        default: nextSt_r = 4'bxxxx;
    endcase

  // Define the current state sequential circuit.
  always @(negedge nRst_i, posedge clk_i)
    if (!nRst_i) begin
      currSt_r <= resetSt_p;  // asynchronous reset
      prevSt_r <= resetSt_p;  // history is reset state too
    end else begin
      currSt_r <= nextSt_r;   // update current state
      prevSt_r <= currSt_r;   // current state delayed 1 clock period

      // Miscellaneous things to do at posedge clock and nRst_i...

      // Store clipped arm period digits (00 to 59 seconds) on the clock
      // edge that moves the FSM from resetSt_p to armSt_p.  This counts
      // on the non-blocking assignment using the value of currSt_r
      // *before* it changes above.
      if (currSt_r == resetSt_p) begin
        armT10s_r <= (armT_i[6:4] > 3'd5) ? 4'd5 : {1'b0, armT_i[6:4]};
        armT1s_r  <= (armT_i[3:0] > 4'd9) ? 4'd9 : armT_i[3:0];
      end

      // Track changes in the React key so the React-to-Display transition
      // can be initiated on the *event* of a key press, not a *state*.
      prev_nReact_r <= nReact_i;

    end

  // *** 5-digit timer and its reset and enable controls ***

  // Reset the timer on a user reset or the Arming-to-React transition.
  assign nRst_timer_w =
    ~(~nRst_i | ((prevSt_r == armSt_p) & (currSt_r == reactSt_p)));
    // I spent a long time trying to figure out why changing this to the
    // following line and eliminating prevSt_r leads to, at the expiration
    // of the arm period, Cnt5Digits resetting to 0 but no transition
    // to the react state.  On the DE2-115 board we stay in the arm state
    // with the counter repeatedly counting from 0 to the arm period.
    // While simulating, looking at all 1s cntrStage signals with and
    // without the change show them to be equal (including no reset at
    // the clock edge when the arm period expires).  Without the change
    // the 1s cntrStage digit goes to the arm period.  With the change
    // it never gets there (although the reset to 0 happens).  REVISIT.
    // ~(~nRst_i | ((currSt_r == armSt_p) & (nextSt_r == reactSt_p)));

  // Enable the timer only in the Arming or React states.
  assign en_timer_w = ((currSt_r == armSt_p) | (currSt_r == reactSt_p));

  Cnt5Digits #(.tc0_p (tc0_p),     // (clks/ms)-1
               .tc1_p (tc1_p),     // (cnt base)-1
               .dig_bits_p (4) )   // 4-bit digits regardless of base
    timer (
      .clk_i      (clk_i),         // 50 MHz master clock
      .nRst_i     (nRst_timer_w),  // active low asynchronous reset
      .enable_i   (en_timer_w),    // active high enable
      .cnt10s_o   (cnt10s_w),      // 4-bit output digits
      .cnt1s_o    (cnt1s_w),
      .cnt100ms_o (cnt100ms_w),
      .cnt10ms_o  (cnt10ms_w),
      .cnt1ms_o   (cnt1ms_w) );

  // *** module outputs ***

  assign reactSt_o = (currSt_r == reactSt_p);

  // Blank the 10s digit when its value is 0 to accomplish leading zero
  // suppression.  Blank the 10 ms and 1 ms digits in all cases except
  // the Display state.
  assign cnt10s_o   =     {cnt10s_w == 4'd0, cnt10s_w};
  assign cnt1s_o    =                 {1'b0, cnt1s_w};
  assign cnt100ms_o =                 {1'b0, cnt100ms_w};
  assign cnt10ms_o  = {currSt_r != dispSt_p, cnt10ms_w};
  assign cnt1ms_o   = {currSt_r != dispSt_p, cnt1ms_w};

endmodule


// ENGR 190 lab 10, Cnt5Digits test bench, count in base 2, 2 clocks/ms
// Rick Nungester 3/12/17

`timescale 1ms / 10us   // times in ms units, to 0.01 ms resolution

module Cnt5Digits_tb;
                    // DUT inputs
  reg clk_50 = 1;   // simulated 50 MHz clock
  reg nRst;         // active-low reset
  reg enable;       // active-high enable (start/stop)

  wire disp_10s;    // DUT outputs, 1 bit for each binary timer digit
  wire disp_1s;
  wire disp_100ms;
  wire disp_10ms;
  wire disp_1ms;

  always                     // clock definition
    #0.25 clk_50 = ~clk_50;  // 2 cycles/ms, +edges at 0.5, 1, 1.5... ms

  Cnt5Digits        // instantiate the Device Under Test
    #(.tc0_p (1),   // 2 clocks per ms, terminal count = 1
      .tc1_p (1) )  // base 2, terminal count = 1 (.dig_bits_p = 1)
      DUT (
        .clk_i      (clk_50),
        .nRst_i     (nRst),
        .enable_i   (enable),
        .cnt10s_o   (disp_10s),  // all 5 digits are 1 bit each
        .cnt1s_o    (disp_1s),
        .cnt100ms_o (disp_100ms),
        .cnt10ms_o  (disp_10ms),
        .cnt1ms_o   (disp_1ms) );

  initial begin            // test bench stimulus and response
    $dumpfile("a.vcd");          // for GTKWave
    $dumpvars(0, Cnt5Digits_tb); // for GTKWave
    $display("Cnt5Digits Simulation: Start");
    enable        = 1'b0;  // @ 0 ms disable counting,
    nRst          = 1'b0;  // and reset count to 0
    #0.25  nRst   = 1'b1;  // @ 0.25 ms (1st clk -edge) release reset,
           enable = 1'b1;  // and start counting on exact ms boundaries
    #34    nRst   = 1'b0;  // @ 34.25 ms clk -edge, mid-00010, asynch reset
    #0.5   nRst   = 1'b1;  // @ 34.75 ms release reset, count...
    #1     enable = 1'b0;  // @ 35.75 ms disable mid-00001, stop counting
    #2     enable = 1'b1;  // @ 37.75 ms enable, start counting from 00001
    #2.25                  // @ 40 ms interrupt count 00011 to finish
    $display("Cnt5Digits Simulation: End");
    $finish;
  end

endmodule


// ENGR 190 lab 10, 5-digit timer (0 to 99.999s) on the DE2-115 board
// Rick Nungester, 3/13/17

module Cnt5Digits_top(
  input        CLOCK_50, // 50 MHz clock
  input        KEY3,     // press to stop, release to continue
  input        KEY0,     // press to reset, release to count
  output [6:0] HEX5,     // 10s digit to 7-segment display
  output [6:0] HEX4,     //  1s digit
  output [6:0] HEX3,     // 100 ms digit
  output [6:0] HEX2,     //  10 ms digit
  output [6:0] HEX1 );   //   1 ms digit

  wire [3:0] disp_10s;   // 4-bit count digits, for 5 digits
  wire [3:0] disp_1s;
  wire [3:0] disp_100ms;
  wire [3:0] disp_10ms;
  wire [3:0] disp_1ms;

  Cnt5Digits
    // #(.tc0_p      (28934),    // 12^-3 second update rate
    //   .tc1_p      (11) )      // base 12 timing (.dig_bits_p = 4)
    // #(.tc0_p      (6249999),  // 12^-3 second update rate
    //   .tc1_p      (1),        // base 2 timing
    //   .dig_bits_p (4) )       // 4-bit digit output ports
    timer (
      .clk_i      (CLOCK_50),
      .nRst_i     (KEY0),
      .enable_i   (KEY3),
      .cnt10s_o   (disp_10s),
      .cnt1s_o    (disp_1s),
      .cnt100ms_o (disp_100ms),
      .cnt10ms_o  (disp_10ms),
      .cnt1ms_o   (disp_1ms) );

  hexBlank7seg sec10s_disp (
    .bn_i   ({disp_10s == 4'b0, disp_10s}),  // blank leading zeros
    .disp_o (HEX5) );

  hexBlank7seg sec1s_disp (
    .bn_i   ({1'b0, disp_1s}),
    .disp_o (HEX4) );

  hexBlank7seg sec100ms_disp (
    .bn_i   ({1'b0, disp_100ms}),
    .disp_o (HEX3) );

  hexBlank7seg sec10ms_disp (
    .bn_i   ({1'b0, disp_10ms}),
    .disp_o (HEX2) );

  hexBlank7seg msec_disp (
    .bn_i   ({1'b0, disp_1ms}),
    .disp_o (HEX1) );

endmodule


// ENGR 190 lab 10, 00.000 to 99.999 second timer with start/stop/reset
// Rick Nungester, 3/12/17

module Cnt5Digits #(
  // Parameter settings for timing dd.ddd seconds in base B.  tc1_p = B-1.
  // tc0_p = round(50e6*T) - 1.
  //  B  tc1_p  T (s)   tc0_p   Notes
  // ==  =====  =====  =======  ======
  //  2    1     2^-3  6249999
  //  8    7     8^-3    97655
  // 10    9    10^-3    49999  default
  // 12   11    12^-3    28934
  // 16   15    16^-3    12206
  parameter tc0_p   = 49999,
  parameter tc1_p   = 9,
  // Bits per output digit defaults to the minimum required to count to
  // tc1_p, but can be set to any higher value if desired.
  parameter dig_bits_p = $clog2(tc1_p + 1) ) (
  input  clk_i,                       // 50 MHz Input clock
  input  nRst_i,                      // asynch active low reset
  input  enable_i,                    // synchronous active high enable
  output [dig_bits_p-1:0] cnt10s_o,   // timer output digits
  output [dig_bits_p-1:0] cnt1s_o,
  output [dig_bits_p-1:0] cnt100ms_o,
  output [dig_bits_p-1:0] cnt10ms_o,
  output [dig_bits_p-1:0] cnt1ms_o );

  wire            tc1msCntr_w;   // 1 ms counter at terminal count
  wire en1ms_w,   tc1ms_w;       // 1 ms digit, enable & at terminal count
  wire en10ms_w,  tc10ms_w;      // 10 ms digit
  wire en100ms_w, tc100ms_w;     // 100 ms digit
  wire en1s_w,    tc1s_w;        // 1s digit
  wire en10s_w;                  // 10s digit, enable (at tc unused)

  // Instantiate cntrStage 6 times, 1 to time 1 ms and 5 for each timer
  // digit used to count ms up to 99.999 seconds.  The cascading
  // enable signals make this a synchronous counter by enabling a stage
  // only when all previous stages are at their terminal counts and
  // enable_i is true.

  cntrStage #(.cntr_tc_p (tc0_p))
    time1ms (                     // 1 ms timer
      .clk_i      (clk_i),
      .nRst_i     (nRst_i),
      .enable_i   (enable_i),
      .term_cnt_o (tc1msCntr_w),
      .count_o    () );

  assign en1ms_w = (enable_i & tc1msCntr_w);

  cntrStage #(.cntr_tc_p (tc1_p),
              .count_o_bits_p (dig_bits_p) )
    cnt1ms (                      // 1 ms digit
      .clk_i      (clk_i),
      .nRst_i     (nRst_i),
      .enable_i   (en1ms_w),
      .term_cnt_o (tc1ms_w),
      .count_o    (cnt1ms_o) );

  assign en10ms_w = (en1ms_w & tc1ms_w);

  cntrStage #(.cntr_tc_p (tc1_p),
              .count_o_bits_p (dig_bits_p) )
    cnt10ms (                     // 10 ms digit
      .clk_i      (clk_i),
      .nRst_i     (nRst_i),
      .enable_i   (en10ms_w),
      .term_cnt_o (tc10ms_w),
      .count_o    (cnt10ms_o) );

  assign en100ms_w = (en10ms_w & tc10ms_w);

  cntrStage #(.cntr_tc_p (tc1_p),
              .count_o_bits_p (dig_bits_p) )
    cnt100ms (                    // 100 ms digit
      .clk_i      (clk_i),
      .nRst_i     (nRst_i),
      .enable_i   (en100ms_w),
      .term_cnt_o (tc100ms_w),
      .count_o    (cnt100ms_o) );

  assign en1s_w = (en100ms_w & tc100ms_w);

  cntrStage #(.cntr_tc_p (tc1_p),
              .count_o_bits_p (dig_bits_p) )
    cnt1s (                       // 1s digit
      .clk_i      (clk_i),
      .nRst_i     (nRst_i),
      .enable_i   (en1s_w),
      .term_cnt_o (tc1s_w),
      .count_o    (cnt1s_o) );

  assign en10s_w = (en1s_w & tc1s_w);

  cntrStage #(.cntr_tc_p (tc1_p),
              .count_o_bits_p (dig_bits_p) )
    cnt10s (                      // 10s digit
      .clk_i      (clk_i),
      .nRst_i     (nRst_i),
      .enable_i   (en10s_w),
      .term_cnt_o (),
      .count_o    (cnt10s_o) );

endmodule


// ENGR 190 lab 10, general-purpose cascadable counter stage
// Rick Nungester 3/5/17

module cntrStage
  #(parameter cntr_tc_p = 9,  // counter terminal count
    // The number of bits in the count_o output defaults to the minimum
    // needed to count from 0 to cntr_tc_p.  But leave it possible
    // to be overridden if it is desired to fix the size of the
    // output port as someting *larger* than required.  For example,
    // this allows terminal count 7, that only requires 3 bits to count
    // to, being output on a 4-bit output bus.
    parameter count_o_bits_p = cntr_bits_p) (
  input  clk_i,       // positive edge triggered clock
  input  nRst_i,      // asynchronous active low reset
  input  enable_i,    // synchronous active high enable
  output term_cnt_o,  // set during terminal count
  output [count_o_bits_p-1:0] count_o ); // count value

  // System task clog2 returns the ceiling of the log base 2 of the
  // argument.  Parameter cntr_bits_p is the minimum number of bits
  // needed to count 0 to cntr_tc_p.  It is not intended to be
  // overridden by higher level code, so is a localparam instead of
  // a parameter.
  localparam cntr_bits_p = $clog2(cntr_tc_p + 1);

  reg [cntr_bits_p-1:0] count_r;    // count register

  always @(posedge clk_i, negedge nRst_i) begin
    if (~nRst_i)
      count_r <= 1'b0;              // if reset, set to 0
    else if (enable_i) begin
      if (count_r == cntr_tc_p)
        count_r <= 1'b0;            // enabled & at tc, wrap to 0
      else
        count_r <= count_r + 1'b1;  // enabled & not TC, increment
    end
  end

  // set outputs from registers used in the always block
  assign count_o = count_r;  // count_o might be more bits than count_r
  assign term_cnt_o = (count_r == cntr_tc_p);

endmodule


// ENGR 190 lab 10, hex to 7-segment decoder, with blanking
// Rick Nungester, 3/6/17

module hexBlank7seg (
  input  [4:0] bn_i,      // blank control (1=blank) and 4-bit hex number
  output [6:0] disp_o );  // 7-segment display output control

  //     7-segment display segment number: 6543210
  assign disp_o = (bn_i[4])           ? 7'b1111111 :
                  (bn_i[3:0] == 4'h0) ? 7'b1000000 :
                  (bn_i[3:0] == 4'h1) ? 7'b1111001 :
                  (bn_i[3:0] == 4'h2) ? 7'b0100100 :
                  (bn_i[3:0] == 4'h3) ? 7'b0110000 :
                  (bn_i[3:0] == 4'h4) ? 7'b0011001 :
                  (bn_i[3:0] == 4'h5) ? 7'b0010010 :
                  (bn_i[3:0] == 4'h6) ? 7'b0000010 :
                  (bn_i[3:0] == 4'h7) ? 7'b1111000 :
                  (bn_i[3:0] == 4'h8) ? 7'b0000000 :
                  (bn_i[3:0] == 4'h9) ? 7'b0011000 :
                  (bn_i[3:0] == 4'hA) ? 7'b0001000 :
                  (bn_i[3:0] == 4'hb) ? 7'b0000011 :
                  (bn_i[3:0] == 4'hC) ? 7'b1000110 :
                  (bn_i[3:0] == 4'hd) ? 7'b0100001 :
                  (bn_i[3:0] == 4'hE) ? 7'b0000110 :
                             /* 4'hF */ 7'b0001110;
endmodule
